/* $Id: init.c,v 1.76 2004/05/31 18:18:55 mikpe Exp $
 * Performance-monitoring counters driver.
 * Top-level initialisation code.
 *
 * Copyright (C) 1999-2004  Mikael Pettersson
 */
#include <linux/config.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/perfctr.h>

#include <asm/uaccess.h>

#include "cpumask.h"
#include "virtual.h"
#include "version.h"

struct perfctr_info perfctr_info = {
	.abi_version = PERFCTR_ABI_VERSION,
	.driver_version = VERSION,
};

char *perfctr_cpu_name __initdata;

static int cpus_copy_to_user(const cpumask_t *cpus, struct perfctr_cpu_mask *argp)
{
	const unsigned int k_nrwords = PERFCTR_CPUMASK_NRLONGS*(sizeof(long)/sizeof(int));
	unsigned int u_nrwords;
	unsigned int ui, ki, j;

	if (get_user(u_nrwords, &argp->nrwords))
		return -EFAULT;
	if (put_user(k_nrwords, &argp->nrwords))
		return -EFAULT;
	if (u_nrwords < k_nrwords)
		return -EOVERFLOW;
	for(ui = 0, ki = 0; ki < PERFCTR_CPUMASK_NRLONGS; ++ki) {
		unsigned long mask = cpus_addr(*cpus)[ki];
		for(j = 0; j < sizeof(long)/sizeof(int); ++j) {
			if (put_user((unsigned int)mask, &argp->mask[ui]))
				return -EFAULT;
			++ui;
			mask = (mask >> (8*sizeof(int)-1)) >> 1;
		}
	}
	return 0;
}

asmlinkage long sys_perfctr_info(struct perfctr_info *infop,
				 struct perfctr_cpu_mask *cpusp,
				 struct perfctr_cpu_mask *forbiddenp)
{
	if (infop && copy_to_user(infop, &perfctr_info, sizeof perfctr_info))
		return -EFAULT;
	if (cpusp) {
		cpumask_t cpus = cpu_online_map;
		int err = cpus_copy_to_user(&cpus, cpusp);
		if (err)
			return err;
	}
	if (forbiddenp) {
		cpumask_t cpus = perfctr_cpus_forbidden_mask;
		int err = cpus_copy_to_user(&cpus, forbiddenp);
		if (err)
			return err;
	}
	return 0;
}

int __init perfctr_init(void)
{
	int err;

	err = perfctr_cpu_init();
	if (err) {
		printk(KERN_INFO "perfctr: not supported by this processor\n");
		return err;
	}
	err = vperfctr_init();
	if (err)
		return err;
	printk(KERN_INFO "perfctr: driver %s, cpu type %s at %u kHz\n",
	       perfctr_info.driver_version,
	       perfctr_cpu_name,
	       perfctr_info.cpu_khz);
	return 0;
}

void __exit perfctr_exit(void)
{
	vperfctr_exit();
	perfctr_cpu_exit();
}

module_init(perfctr_init)
module_exit(perfctr_exit)
